# noqa: D100
import re
import sys
import textwrap

from enum import Enum
from pathlib import Path
from typing import Annotated, TYPE_CHECKING

import typer

from dbrownell_Common.Streams.DoneManager import DoneManager, Flags as DoneManagerFlags
from semantic_version import Version as SemVer
from typer.core import TyperGroup

from dbrownell_ToolsDirectory import __version__
from dbrownell_ToolsDirectory.CreateShellCommands import CreateShellCommands
from dbrownell_ToolsDirectory.ManifestGenerator import GenerateManifest, WriteManifestYaml
from dbrownell_ToolsDirectory.Shell.BashCommandVisitor import BashCommandVisitor
from dbrownell_ToolsDirectory.Shell.BatchCommandVisitor import BatchCommandVisitor
from dbrownell_ToolsDirectory.Shell.PowerShellCommandVisitor import PowerShellCommandVisitor
from dbrownell_ToolsDirectory import ToolInfo

if TYPE_CHECKING:
    from dbrownell_ToolsDirectory.Shell.Commands import Command


# ----------------------------------------------------------------------
class NaturalOrderGrouper(TyperGroup):  # noqa: D101
    # ----------------------------------------------------------------------
    def list_commands(self, *args, **kwargs) -> list[str]:  # noqa: ARG002, D102
        return list(self.commands.keys())  # pragma: no cover


# ----------------------------------------------------------------------
app = typer.Typer(
    cls=NaturalOrderGrouper,
    help=__doc__,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=False,
)


# ----------------------------------------------------------------------
class OutputType(str, Enum):
    """Specify the type of output to generate."""

    Bash = "bash"
    Batch = "batch"
    PowerShell = "powershell"


# ----------------------------------------------------------------------
def _CreateHelp() -> str:
    return textwrap.dedent(
        r"""Create statements that add tools to the path.

        Tools should be organized by:

        <tool_directory>/
        └── <tool_name>/
            └── \[<version>/]
                └── \[<operating_system:{operating_systems}>/]
                    └── \[<architecture:{architectures}>/]
                        └── \[bin/]

        Each of these examples are supported:

        Tools/
        ├── Tool1/
        ├── Tool2/bin/
        ├── Tool3/
        │   └── 1.0.0/
        ├── Tool3/
        │   └── 1.0.0/
        │       └── bin/
        ├── Tool4/
        │   └── v1.0.0/
        ├── Tool5/
        │   └── 1.0.0/
        │       └── Linux/
        ├── Tool6/
        │   └── 1.0.0/
        │       └── Linux/
        │           └── bin/
        ├── Tool7/
        │   └── 1.0.0/
        │       └── Linux/
        │           └── x64/
        └── Tool8/
            └── 1.0.0/
                └── Linux/
                    └── x64/
                        └── bin/

        """,
    ).format(
        operating_systems=" | ".join(ToolInfo.OperatingSystemType.StringMap().keys()),
        architectures=" | ".join(ToolInfo.ArchitectureType.StringMap().keys()),
    )


# ----------------------------------------------------------------------
@app.command("activate", help=_CreateHelp(), no_args_is_help=True)
def Activate(  # noqa: D103
    output_filename: Annotated[
        Path,
        typer.Argument(dir_okay=False, resolve_path=True, path_type=Path, help="Path to the output file."),  # ty: ignore[no-matching-overload]
    ],
    output_type: Annotated[
        OutputType,
        typer.Argument(help="The type of output to generate."),
    ],
    tool_directory: Annotated[
        Path,
        typer.Argument(  # ty: ignore[no-matching-overload]
            exists=True,
            file_okay=False,
            resolve_path=True,
            path_type=Path,
            help="Path to the tool directory.",
        ),
    ],
    include: Annotated[
        list[str] | None,
        typer.Option(
            "--include", "-i", help="Name of a tool to include; all tools found are included by default."
        ),
    ] = None,
    exclude: Annotated[
        list[str] | None,
        typer.Option("--exclude", "-e", help="Name of a tool to exclude; no tools are excluded by default."),
    ] = None,
    tool_version: Annotated[
        list[str] | None,
        typer.Option(
            "--tool-version",
            "-tv",
            help="Specify the specific version of a tool to use; the latest version is used by default.",
        ),
    ] = None,
    no_generic_operating_system: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--no-generic-operating-system",
            "-ngos",
            help="Do not consider 'Generic' operating system versions when calculating the correct version.",
        ),
    ] = False,
    no_generic_architecture: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--no-generic-architecture",
            "-ngarch",
            help="Do not consider 'Generic' architecture versions when calculating the correct version.",
        ),
    ] = False,
    verbose: Annotated[  # noqa: FBT002
        bool,
        typer.Option("--verbose", help="Write verbose information to the terminal."),
    ] = False,
    debug: Annotated[  # noqa: FBT002
        bool,
        typer.Option("--debug", help="Write debug information to the terminal."),
    ] = False,
) -> None:
    with DoneManager.CreateCommandLine(
        flags=DoneManagerFlags.Create(verbose=verbose, debug=debug),
    ) as dm:
        if output_filename.is_file():
            output_filename.unlink()

        # Get the tools
        tool_infos = ToolInfo.GetAllToolInfos(
            dm,
            tool_directory,
            set(include or []),
            set(exclude or []),
            _ExtractToolVersions(dm, tool_version or []),
            ToolInfo.OperatingSystemType.Calculate(),
            ToolInfo.ArchitectureType.Calculate(),
            allow_generic_operating_systems=not no_generic_operating_system,
            allow_generic_architectures=not no_generic_architecture,
        )

        if dm.result != 0:
            return

        if not tool_infos:
            dm.WriteError(f"No tools were found in '{tool_directory}'.\n")
            return

        # Display the tools
        if dm.is_verbose:
            for tool_info in tool_infos:
                header = tool_info.name

                if tool_info.version is not None:
                    header += f" (v{tool_info.version})"

                dm.WriteLine(
                    textwrap.dedent(
                        """\
                        {header}
                        {sep}
                        Root:              {root}
                        Versioned:         {versioned}
                        Binary:            {binary}
                        Operating System:  {os}
                        Architecture:      {arch}

                        """,
                    ).format(
                        header=header,
                        sep="-" * max(len(header), 17),
                        root=tool_info.root_directory,
                        versioned=tool_info.versioned_directory,
                        binary=tool_info.binary_directory,
                        os=tool_info.operating_system.name
                        if isinstance(tool_info.operating_system, ToolInfo.OperatingSystemType)
                        else tool_info.operating_system,
                        arch=tool_info.architecture.name
                        if isinstance(tool_info.architecture, ToolInfo.ArchitectureType)
                        else tool_info.architecture,
                    ),
                )

        # Create the shell statements
        commands: list[Command] = CreateShellCommands(dm, tool_infos)

        # Write the output
        if output_type == OutputType.Bash:
            command_visitor = BashCommandVisitor()
        elif output_type == OutputType.Batch:
            command_visitor = BatchCommandVisitor()
        elif output_type == OutputType.PowerShell:
            command_visitor = PowerShellCommandVisitor()
        else:
            assert False, output_type  # noqa: B011, PT015  # pragma: no cover

        output_filename.parent.mkdir(parents=True, exist_ok=True)
        with output_filename.open("w", encoding="utf-8") as f:
            for command in commands:
                result = command_visitor.Accept(command)

                if isinstance(result, str):
                    f.write(result)


# ----------------------------------------------------------------------
@app.command("manifest", no_args_is_help=True)
def Manifest(
    tool_directory: Annotated[
        Path,
        typer.Argument(  # ty: ignore[no-matching-overload]
            exists=True,
            file_okay=False,
            resolve_path=True,
            path_type=Path,
            help="Path to the tool directory.",
        ),
    ],
    output_filename: Annotated[
        Path,
        typer.Argument(  # ty: ignore[no-matching-overload]
            dir_okay=False,
            resolve_path=True,
            path_type=Path,
            help="Path to the output YAML file.",
        ),
    ],
    include: Annotated[
        list[str] | None,
        typer.Option(
            "--include", "-i", help="Name of a tool to include; all tools found are included by default."
        ),
    ] = None,
    exclude: Annotated[
        list[str] | None,
        typer.Option("--exclude", "-e", help="Name of a tool to exclude; no tools are excluded by default."),
    ] = None,
    verbose: Annotated[  # noqa: FBT002
        bool,
        typer.Option("--verbose", help="Write verbose information to the terminal."),
    ] = False,
    debug: Annotated[  # noqa: FBT002
        bool,
        typer.Option("--debug", help="Write debug information to the terminal."),
    ] = False,
) -> None:
    """Generate a YAML manifest of all tools in a directory."""

    with DoneManager.CreateCommandLine(
        flags=DoneManagerFlags.Create(verbose=verbose, debug=debug),
    ) as dm:
        manifest = GenerateManifest(
            dm,
            tool_directory,
            include_tools=set(include) if include else None,
            exclude_tools=set(exclude) if exclude else None,
        )

        WriteManifestYaml(manifest, output_filename)


# ----------------------------------------------------------------------
@app.command("version", no_args_is_help=False)
def Version() -> None:
    """Display the version of this tool."""
    sys.stdout.write(f"{__version__}\n")


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _ExtractToolVersions(
    dm: DoneManager,
    args: list[str],
) -> dict[str, SemVer]:
    kvp_regex = re.compile(
        r"""
        ^                                   # Beginning of string
        (?P<tool_name>.+?)                  # Tool name
        \s*(?<!\\)\=\s*                     # Equal sign (not escaped)
        v?                                  # Optional 'v' prefix
        (?P<version>.+?)                    # Version
        $                                   # End of string
        """,
        re.VERBOSE,
    )

    results: dict[str, SemVer] = {}

    for arg in args:
        match = kvp_regex.match(arg)
        if not match:
            dm.WriteError(f"The tool version '{arg}' is not a valid command line argument.\n")
            continue

        tool_name = match.group("tool_name")
        version_str = match.group("version")

        try:
            version = SemVer.coerce(version_str)
        except ValueError:
            dm.WriteError(f"The version '{version_str}' for the tool '{tool_name}' is invalid.\n")
            continue

        results[tool_name] = version

    return results


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
if __name__ == "__main__":
    app()  # pragma: no cover



# noqa: D100
import textwrap

from dbrownell_Common.Types import override

from dbrownell_ToolsDirectory.Shell.Commands import (
    Message,
    Call,
    Execute,
    Set,
    Augment,
    Exit,
    ExitOnError,
    EchoOff,
    PersistError,
    PushDirectory,
    PopDirectory,
    Raw,
)
from dbrownell_ToolsDirectory.Shell.CommandVisitor import CommandVisitor


# ----------------------------------------------------------------------
class BashCommandVisitor(CommandVisitor):
    """Command visitor for Linux bash scripts."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__()

        self._message_substitution_lookup: dict[str, str] = {
            "$": r"\$",
            '"': r"\"",
            "`": r"\\\`",
        }

    # ----------------------------------------------------------------------
    @override
    def OnMessage(  # noqa: D102
        self,
        command: Message,
    ) -> str | None:
        output: list[str] = []

        for command_line in command.value.split("\n"):
            if not command_line.strip():
                output.append('echo ""')
                continue

            line = command_line

            for source, dest in self._message_substitution_lookup.items():
                line = line.replace(source, dest)

            output.append(f'echo "{line}"')

        return " && ".join(output) + "\n"

    # ----------------------------------------------------------------------
    @override
    def OnCall(  # noqa: D102
        self,
        command: Call,
    ) -> str | None:
        result = f"source {command.command_line}\n"
        if command.exit_on_error:
            exit_on_error_result = self.Accept(
                ExitOnError(use_return_statement=command.exit_via_return_statement)
            )

            if exit_on_error_result:
                result += exit_on_error_result

        return result

    # ----------------------------------------------------------------------
    @override
    def OnExecute(  # noqa: D102
        self,
        command: Execute,
    ) -> str | None:
        result = command.command_line + "\n"

        if command.exit_on_error:
            exit_on_error_result = self.Accept(
                ExitOnError(use_return_statement=command.exit_via_return_statement)
            )

            if exit_on_error_result:
                result += exit_on_error_result

        return result

    # ----------------------------------------------------------------------
    @override
    def OnSet(  # noqa: D102
        self,
        command: Set,
    ) -> str | None:
        if command.value_or_values is None:
            return f"unset {command.name}\n"

        values = ":".join(command.EnumValues())

        values = values.removeprefix('"')
        values = values.removesuffix('"')

        return f'export {command.name}="{values}"\n'

    # ----------------------------------------------------------------------
    @override
    def OnAugment(  # noqa: D102
        self,
        command: Augment,
    ) -> str | None:
        if command.append_values:
            add_statement_template = f"{{value}}:${{{{{command.name}}}}}"
        else:
            add_statement_template = f"${{{{{command.name}}}}}:{{value}}"

        add_statement_template = f'export {command.name}="{add_statement_template}"'

        # Apply true to ensure that "$?" isn't set to 1
        statement_template = textwrap.dedent(
            f"""\
            [[ ":${{{{{command.name}}}}}:" != *":{{value}}:"* ]] && {add_statement_template}
            true
            """,
        )

        statements: list[str] = [statement_template.format(value=value) for value in command.EnumValues()]

        return "".join(statements)

    # ----------------------------------------------------------------------
    @override
    def OnExit(  # noqa: D102
        self,
        command: Exit,
    ) -> str | None:
        return textwrap.dedent(
            """\
            {success}
            {error}
            return {return_code}
            """,
        ).format(
            success=textwrap.dedent(
                """\
                if [[ $? -eq 0 ]]; then
                    read -p "Press [Enter] to continue"
                fi
                """,
            ).rstrip()
            if command.pause_on_success
            else "",
            error=textwrap.dedent(
                """\
                if [[ $? -ne 0 ]]; then
                    read -p "Press [Enter] to continue"
                fi
                """,
            ).rstrip()
            if command.pause_on_error
            else "",
            return_code=command.return_code or 0,
        )

    # ----------------------------------------------------------------------
    @override
    def OnExitOnError(  # noqa: D102
        self,
        command: ExitOnError,
    ) -> str | None:
        variable_name = f"${command.variable_name}" if command.variable_name else "$?"

        return textwrap.dedent(
            f"""\
            error_code={variable_name}
            if [[ $error_code -ne 0 ]]; then
                {"return" if command.use_return_statement else "exit"} {command.return_code or "$error_code"}
            fi
            """,
        )

    # ----------------------------------------------------------------------
    @override
    def OnEchoOff(  # noqa: D102
        self,
        command: EchoOff,  # noqa: ARG002
    ) -> str | None:
        return textwrap.dedent(
            """\
            set +v
            set +x

            """,
        )

    # ----------------------------------------------------------------------
    @override
    def OnPersistError(  # noqa: D102
        self,
        command: PersistError,
    ) -> str | None:
        return f"{command.variable_name}=$?\n"

    # ----------------------------------------------------------------------
    @override
    def OnPushDirectory(  # noqa: D102
        self,
        command: PushDirectory,
    ) -> str | None:
        if command.value is None:
            directory = """$( cd "$( dirname "${BASH_SOURCE[0]}" )" > /dev/null 2>&1 && pwd )"""
        else:
            directory = f'"{command.value.as_posix()}"'

        return f"pushd {directory} > /dev/null\n"

    # ----------------------------------------------------------------------
    @override
    def OnPopDirectory(  # noqa: D102
        self,
        command: PopDirectory,  # noqa: ARG002
    ) -> str | None:
        return "popd > /dev/null\n"

    # ----------------------------------------------------------------------
    @override
    def OnRaw(  # noqa: D102
        self,
        command: Raw,
    ) -> str | None:
        return command.value.removesuffix("\n") + "\n"



# noqa: D100
import platform
import sys

from dataclasses import dataclass
from enum import auto, Enum
from typing import Literal, TYPE_CHECKING

from dbrownell_Common.InflectEx import inflect
from semantic_version import Version as SemVer

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

    from dbrownell_Common.Streams.DoneManager import DoneManager


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
class OperatingSystemType(Enum):
    """Specify the operating system."""

    Linux = auto()
    MacOS = auto()
    Windows = auto()

    # ----------------------------------------------------------------------
    @classmethod
    def Calculate(cls) -> OperatingSystemType:
        """Calculate the current operating system."""

        platform_str = sys.platform.lower()

        if platform_str.startswith("linux"):
            return OperatingSystemType.Linux

        if platform_str.startswith("darwin"):
            return OperatingSystemType.MacOS

        if platform_str.startswith("win32"):
            return OperatingSystemType.Windows

        msg = f"Unsupported platform: '{platform_str}'"
        raise Exception(msg)

    # ----------------------------------------------------------------------
    @classmethod
    def StringMap(cls) -> dict[str, OperatingSystemType | Literal["Generic"]]:
        """Return strings that represent operating systems."""

        results: dict[str, OperatingSystemType | Literal["Generic"]] = {e.name: e for e in cls}

        results["Generic"] = "Generic"

        return results


# ----------------------------------------------------------------------
class ArchitectureType(Enum):
    """Specify the system architecture."""

    x64 = auto()
    x86 = auto()
    ARM64 = auto()
    ARM = auto()

    # ----------------------------------------------------------------------
    @classmethod
    def Calculate(cls) -> ArchitectureType:
        """Calculate the current system architecture."""

        arch = platform.machine().lower()

        if arch == "arm64":
            return ArchitectureType.ARM64
        if arch == "amd":
            return ArchitectureType.ARM

        if arch not in ["x86_64", "amd64"]:
            msg = f"Unsupported architecture: '{arch}'"
            raise Exception(msg)

        return ArchitectureType.x64 if sys.maxsize > 2**32 else ArchitectureType.x86

    # ----------------------------------------------------------------------
    @classmethod
    def StringMap(cls) -> dict[str, ArchitectureType | Literal["Generic"]]:
        """Return strings that represent architectures."""

        results: dict[str, ArchitectureType | Literal["Generic"]] = {e.name: e for e in cls}

        results["Generic"] = "Generic"

        return results


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class ToolInfo:
    """Information about a tool."""

    name: str
    """Name of the tool."""

    version: SemVer | None
    """Version of the tool, if a versioned directory was encountered."""

    operating_system: OperatingSystemType | Literal["Generic"] | None
    """Operating system of the tool, if an OS-specific directory was encountered."""

    architecture: ArchitectureType | Literal["Generic"] | None
    """Architecture of the tool, if an architecture-specific directory was encountered."""

    root_directory: Path
    """Root of the tool; all versioned directories are subdirectories of this."""

    versioned_directory: Path
    """Root of a specific version of the tool."""

    binary_directory: Path
    """Binary directory of a specific version of the tool."""

    # ----------------------------------------------------------------------
    def GeneratePotentialEnvFiles(self, file_extension: str = ".env") -> Generator[Path]:
        """Generate potential filenames for environment files, ordered from the least- to most-specific."""

        # ----------------------------------------------------------------------
        def ApplyStandardSuffixes(potential_suffixes: list[str]) -> None:
            assert not potential_suffixes
            potential_suffixes.append("")

        # ----------------------------------------------------------------------
        def ApplyVersionSuffixes(potential_suffixes: list[str]) -> None:
            if self.version is None:
                return

            potential_suffixes.append(f"-v{self.version}")

        # ----------------------------------------------------------------------
        def ApplyOperatingSystemSuffixes(potential_suffixes: list[str]) -> None:
            if self.operating_system is None:
                return

            suffix = (
                self.operating_system.name
                if isinstance(self.operating_system, OperatingSystemType)
                else self.operating_system
            )

            potential_suffixes.append(f"-{suffix}")
            potential_suffixes += [
                f"{potential_suffix}-{suffix}" for potential_suffix in potential_suffixes[1:-1]
            ]

        # ----------------------------------------------------------------------
        def ApplyArchitectureSuffixes(potential_suffixes: list[str]) -> None:
            if self.architecture is None:
                return

            suffix = (
                self.architecture.name
                if isinstance(self.architecture, ArchitectureType)
                else self.architecture
            )

            potential_suffixes.append(f"-{suffix}")
            potential_suffixes += [
                f"{potential_suffix}-{suffix}" for potential_suffix in potential_suffixes[1:-1]
            ]

        # ----------------------------------------------------------------------

        # Get a list of directories between the root and the versioned directory. Append "fake_file"
        # so that versioned_directory is included in the parents (without "fake_file", the implementation
        # thinks that versioned_directory is a file and not a directory).
        relative_paths = list(
            reversed((self.versioned_directory / "fake_file").relative_to(self.root_directory).parents)
        )

        # At most, there can be relative paths for:
        # 1) Root
        # 2) Version
        # 3) Operating System
        # 4) Architecture
        assert len(relative_paths) <= 4, relative_paths  # noqa: PLR2004

        # Root
        potential_suffixes: list[str] = []

        ApplyStandardSuffixes(potential_suffixes)
        ApplyVersionSuffixes(potential_suffixes)
        ApplyOperatingSystemSuffixes(potential_suffixes)
        ApplyArchitectureSuffixes(potential_suffixes)

        relative_path_offset = 0
        root = self.root_directory / relative_paths[relative_path_offset]

        yield from [root / f"{self.name}{suffix}{file_extension}" for suffix in potential_suffixes]

        # Version
        if self.version is not None:
            potential_suffixes = []

            ApplyStandardSuffixes(potential_suffixes)
            ApplyOperatingSystemSuffixes(potential_suffixes)
            ApplyArchitectureSuffixes(potential_suffixes)

            relative_path_offset += 1
            root = self.root_directory / relative_paths[relative_path_offset]

            yield from [root / f"{self.name}{suffix}{file_extension}" for suffix in potential_suffixes]

        # Operating System
        if self.operating_system is not None:
            potential_suffixes = []

            ApplyStandardSuffixes(potential_suffixes)
            ApplyArchitectureSuffixes(potential_suffixes)

            relative_path_offset += 1
            root = self.root_directory / relative_paths[relative_path_offset]

            yield from [root / f"{self.name}{suffix}{file_extension}" for suffix in potential_suffixes]

        # Architecture
        if self.architecture is not None:
            potential_suffixes = []

            ApplyStandardSuffixes(potential_suffixes)

            relative_path_offset += 1
            root = self.root_directory / relative_paths[relative_path_offset]

            yield root / f"{self.name}{file_extension}"


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def GetAllToolInfos(
    dm: DoneManager,
    tools_directory: Path,
    include_tools: set[str],
    exclude_tools: set[str],
    tool_versions: dict[str, SemVer],
    operating_system: OperatingSystemType,
    architecture: ArchitectureType,
    *,
    allow_generic_operating_systems: bool = True,
    allow_generic_architectures: bool = True,
) -> list[ToolInfo]:
    """Return information about tools in a specified directory."""

    results: list[ToolInfo] = []

    with dm.Nested(
        f"Parsing tools in '{tools_directory}'...",
        lambda: "{} found".format(inflect.no("tool", len(results))),
    ) as tool_dm:
        for tool_directory in tools_directory.iterdir():
            if not tool_directory.is_dir():
                continue

            tool_name = tool_directory.name

            if tool_name in exclude_tools:
                tool_dm.WriteVerbose(f"'{tool_name}' has been explicitly excluded.\n")
                continue

            if include_tools and tool_name not in include_tools:
                tool_dm.WriteVerbose(f"'{tool_name}' has not been explicitly included.\n")
                continue

            try:
                tool_infos = list(
                    GenerateToolInfos(
                        tool_directory,
                        tool_versions.get(tool_name) or "latest",
                        operating_system,
                        architecture,
                        allow_generic_operating_system=allow_generic_operating_systems,
                        allow_generic_architecture=allow_generic_architectures,
                    ),
                )
            except ValueError as ex:
                tool_dm.WriteError(str(ex))
                continue

            if not tool_infos:
                tool_dm.WriteError(
                    f"No valid configurations found for the tool '{tool_name}' in '{tool_directory}'.\n"
                )
                continue

            assert len(tool_infos) == 1, tool_infos
            results.append(tool_infos[0])

    return results


# ----------------------------------------------------------------------
def GenerateToolInfos(
    tool_dir: Path,
    version_filter: SemVer | Literal["latest"] | None,
    operating_system_filter: OperatingSystemType | None,
    architecture_filter: ArchitectureType | None,
    *,
    allow_generic_operating_system: bool = True,
    allow_generic_architecture: bool = True,
) -> Generator[ToolInfo]:
    """Generate one or more ToolInfo objects for a specified tool directory."""

    tool_name = tool_dir.name

    for versioned_dir, version in _GenerateVersionedDirs(
        tool_name,
        tool_dir,
        version_filter,
    ):
        for operating_system_dir, operating_system in _GenerateOperatingSystemDirs(
            tool_name,
            versioned_dir,
            operating_system_filter,
            allow_generic_operating_system=allow_generic_operating_system,
        ):
            for architecture_dir, architecture in _GenerateArchitectureDirs(
                tool_name,
                operating_system_dir,
                architecture_filter,
                allow_generic_architecture=allow_generic_architecture,
            ):
                potential_bin_dir = architecture_dir / "bin"
                bin_dir = potential_bin_dir if potential_bin_dir.is_dir() else architecture_dir

                yield ToolInfo(
                    tool_name,
                    version,
                    operating_system,
                    architecture,
                    tool_dir,
                    architecture_dir,
                    bin_dir,
                )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GenerateVersionedDirs(
    tool_name: str,
    working_dir: Path,
    dir_filter: SemVer | Literal["latest"] | None,
) -> Generator[tuple[Path, SemVer | None]]:
    versioned_infos: list[tuple[SemVer, Path]] = []

    for potential_dir in working_dir.iterdir():
        if not potential_dir.is_dir():
            continue

        try:
            version = SemVer.coerce(potential_dir.name.removeprefix("v"))
            versioned_infos.append((version, potential_dir))
        except ValueError:
            # This is an indication that the directory is not a version directory
            continue

    if not versioned_infos:
        yield working_dir, None
        return

    versioned_infos.sort(key=lambda x: x[0], reverse=True)

    if dir_filter is not None:
        versioned_infos_dict = dict(versioned_infos)

        if isinstance(dir_filter, SemVer):
            desired_version = dir_filter
        elif isinstance(dir_filter, str) and dir_filter == "latest":
            desired_version = next(iter(versioned_infos_dict))
        else:
            assert False, dir_filter  # noqa: B011, PT015  # pragma: no cover

        updated_working_dir = versioned_infos_dict.get(desired_version)
        if updated_working_dir is None:
            msg = f"No directory found for version '{desired_version}' for the tool '{tool_name}' in '{working_dir}'."
            raise ValueError(msg)

        yield updated_working_dir, desired_version

    else:
        for versioned_info in versioned_infos:
            yield versioned_info[1], versioned_info[0]


# ----------------------------------------------------------------------
def _GenerateOperatingSystemDirs(
    tool_name: str,
    working_dir: Path,
    dir_filter: OperatingSystemType | None,
    *,
    allow_generic_operating_system: bool,
) -> Generator[tuple[Path, OperatingSystemType | Literal["Generic"] | None]]:
    os_string_map = OperatingSystemType.StringMap()
    os_infos: dict[str, tuple[Path, OperatingSystemType | Literal["Generic"]]] = {}

    for potential_dir in working_dir.iterdir():
        if not potential_dir.is_dir():
            continue

        potential_mapped_type = os_string_map.get(potential_dir.name)
        if potential_mapped_type is not None:
            os_infos[potential_dir.name] = (potential_dir, potential_mapped_type)

    if not os_infos:
        yield working_dir, None
        return

    if dir_filter is not None:
        os_info = os_infos.get(dir_filter.name)

        if os_info is None:
            if allow_generic_operating_system:
                os_info = os_infos.get("Generic")

            if os_info is None:
                msg = f"No directory found for '{dir_filter.name}' for the tool '{tool_name}' in '{working_dir}'."
                raise ValueError(msg)

        yield os_info

    else:
        yield from os_infos.values()


# ----------------------------------------------------------------------
def _GenerateArchitectureDirs(
    tool_name: str,
    working_dir: Path,
    dir_filter: ArchitectureType | None,
    *,
    allow_generic_architecture: bool,
) -> Generator[tuple[Path, ArchitectureType | Literal["Generic"] | None]]:
    arch_string_map = ArchitectureType.StringMap()
    arch_infos: dict[str, tuple[Path, ArchitectureType | Literal["Generic"]]] = {}

    for potential_dir in working_dir.iterdir():
        if not potential_dir.is_dir():
            continue

        potential_mapped_type = arch_string_map.get(potential_dir.name)
        if potential_mapped_type is not None:
            arch_infos[potential_dir.name] = (potential_dir, potential_mapped_type)

    if not arch_infos:
        yield working_dir, None
        return

    if dir_filter is not None:
        arch_info = arch_infos.get(dir_filter.name)

        if arch_info is None:
            if allow_generic_architecture:
                arch_info = arch_infos.get("Generic")

            if arch_info is None:
                msg = f"No directory found for '{dir_filter.name}' for the tool '{tool_name}' in '{working_dir}'."
                raise ValueError(msg)

        yield arch_info

    else:
        yield from arch_infos.values()



# noqa: D100
from dataclasses import dataclass, field
from typing import Literal, TYPE_CHECKING

import yaml

from dbrownell_Common.InflectEx import inflect

from dbrownell_ToolsDirectory.ToolInfo import (
    ArchitectureType,
    GenerateToolInfos,
    OperatingSystemType,
    ToolInfo,
)

if TYPE_CHECKING:
    from pathlib import Path

    from dbrownell_Common.Streams.DoneManager import DoneManager
    from semantic_version import Version as SemVer


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
@dataclass(frozen=True)
class ToolConfiguration:
    """A specific configuration of a tool (version + OS + architecture combination)."""

    version: SemVer | None
    """Version of the tool, if a versioned directory was encountered."""

    operating_system: OperatingSystemType | Literal["Generic"] | None
    """Operating system of the tool, if an OS-specific directory was encountered."""

    architecture: ArchitectureType | Literal["Generic"] | None
    """Architecture of the tool, if an architecture-specific directory was encountered."""

    versioned_directory: Path
    """Root of a specific version of the tool, relative to tools_directory."""

    binary_directory: Path
    """Binary directory of a specific version of the tool, relative to tools_directory."""

    env_files: list[Path] = field(default_factory=list)
    """Environment files that exist on disk, relative to tools_directory."""


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class ToolManifestEntry:
    """A tool with all its configurations."""

    name: str
    """Name of the tool."""

    configurations: list[ToolConfiguration] = field(default_factory=list)
    """All configurations (version + OS + architecture combinations), sorted by version in descending order."""


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class ToolsManifest:
    """Complete manifest of all tools in a directory."""

    tools: list[ToolManifestEntry] = field(default_factory=list)
    """All tools in the directory."""


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def GenerateManifest(
    dm: DoneManager,
    tools_directory: Path,
    include_tools: set[str] | None = None,
    exclude_tools: set[str] | None = None,
) -> ToolsManifest:
    """Generate a manifest of all tools in a directory."""

    manifest_entries: list[ToolManifestEntry] = []

    with dm.Nested(
        f"Generating manifest for '{tools_directory}'...",
        lambda: "{} found".format(inflect.no("tool", len(manifest_entries))),
    ) as manifest_dm:
        include_tools = include_tools or set()
        exclude_tools = exclude_tools or set()

        # Collect ToolInfo objects by tool name
        tool_groups: dict[str, list[ToolInfo]] = {}

        for tool_directory in tools_directory.iterdir():
            if not tool_directory.is_dir():
                continue

            tool_name = tool_directory.name

            if tool_name in exclude_tools:
                manifest_dm.WriteVerbose(f"'{tool_name}' has been explicitly excluded.\n")
                continue

            if include_tools and tool_name not in include_tools:
                manifest_dm.WriteVerbose(f"'{tool_name}' has not been explicitly included.\n")
                continue

            # Generate all ToolInfo objects without filters
            tool_infos = list(
                GenerateToolInfos(
                    tool_directory,
                    version_filter=None,
                    operating_system_filter=None,
                    architecture_filter=None,
                ),
            )

            if not tool_infos:
                manifest_dm.WriteVerbose(f"No configurations found for '{tool_name}'.\n")
                continue

            tool_groups[tool_name] = tool_infos

        # Build the manifest
        for tool_name in sorted(tool_groups.keys()):
            tool_infos = tool_groups[tool_name]

            tool_configurations: list[ToolConfiguration] = []

            for tool_info in tool_infos:
                # Find existing env files (relative to tools_directory)
                existing_env_files: list[Path] = [
                    env_file.relative_to(tools_directory)
                    for env_file in tool_info.GeneratePotentialEnvFiles()
                    if env_file.is_file()
                ]

                tool_configurations.append(
                    ToolConfiguration(
                        version=tool_info.version,
                        operating_system=tool_info.operating_system,
                        architecture=tool_info.architecture,
                        versioned_directory=tool_info.versioned_directory.relative_to(tools_directory),
                        binary_directory=tool_info.binary_directory.relative_to(tools_directory),
                        env_files=existing_env_files,
                    ),
                )

            manifest_entries.append(
                ToolManifestEntry(
                    name=tool_name,
                    configurations=tool_configurations,
                ),
            )

    return ToolsManifest(tools=manifest_entries)


# ----------------------------------------------------------------------
def WriteManifestYaml(
    manifest: ToolsManifest,
    output_path: Path,
) -> None:
    """Write a manifest to a YAML file."""

    yaml_data = _ManifestToDict(manifest)

    output_path.parent.mkdir(parents=True, exist_ok=True)

    with output_path.open("w", encoding="utf-8") as f:
        yaml.dump(yaml_data, f, default_flow_style=False, allow_unicode=True, sort_keys=False)


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _ManifestToDict(manifest: ToolsManifest) -> dict:
    """Convert a ToolsManifest to a dictionary suitable for YAML serialization."""

    return {
        "tools": [_ToolManifestEntryToDict(entry) for entry in manifest.tools],
    }


# ----------------------------------------------------------------------
def _ToolManifestEntryToDict(entry: ToolManifestEntry) -> dict:
    """Convert a ToolManifestEntry to a dictionary."""

    return {
        "name": entry.name,
        "configurations": [_ToolConfigurationToDict(config) for config in entry.configurations],
    }


# ----------------------------------------------------------------------
def _ToolConfigurationToDict(config: ToolConfiguration) -> dict:
    """Convert a ToolConfiguration to a dictionary."""

    os_value: str | None

    if config.operating_system is None:
        os_value = None
    elif isinstance(config.operating_system, OperatingSystemType):
        os_value = config.operating_system.name
    elif isinstance(config.operating_system, str) and config.operating_system == "Generic":
        os_value = config.operating_system
    else:
        assert False, config.operating_system  # noqa: B011, PT015  # pragma: no cover

    arch_value: str | None

    if config.architecture is None:
        arch_value = None
    elif isinstance(config.architecture, ArchitectureType):
        arch_value = config.architecture.name
    elif isinstance(config.architecture, str) and config.architecture == "Generic":
        arch_value = config.architecture
    else:
        assert False, config.architecture  # noqa: B011, PT015  # pragma: no cover

    return {
        "version": str(config.version) if config.version is not None else None,
        "operating_system": os_value,
        "architecture": arch_value,
        "versioned_directory": str(config.versioned_directory.as_posix()),
        "binary_directory": str(config.binary_directory.as_posix()),
        "env_files": [str(p.as_posix()) for p in config.env_files],
    }



# noqa: D100
from abc import abstractmethod, ABC

from dbrownell_ToolsDirectory.Shell.Commands import (
    Command,
    Message,
    Call,
    Execute,
    Set,
    Augment,
    Exit,
    ExitOnError,
    EchoOff,
    PersistError,
    PushDirectory,
    PopDirectory,
    Raw,
)


# ----------------------------------------------------------------------
class CommandVisitor(ABC):
    """Abstract base class for a visitor that processes shell commands."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        self._dispatcher = {
            Message: self.OnMessage,
            Call: self.OnCall,
            Execute: self.OnExecute,
            Set: self.OnSet,
            Augment: self.OnAugment,
            Exit: self.OnExit,
            ExitOnError: self.OnExitOnError,
            EchoOff: self.OnEchoOff,
            PersistError: self.OnPersistError,
            PushDirectory: self.OnPushDirectory,
            PopDirectory: self.OnPopDirectory,
            Raw: self.OnRaw,
        }

    # ----------------------------------------------------------------------
    def Accept(self, command: Command) -> str | None:
        """Accept a command and dispatch it to the appropriate handler."""

        handler = self._dispatcher.get(type(command))
        if handler is None:
            msg = f"No handler for command type: '{command.__class__.__name__}'"
            raise TypeError(msg)

        return handler(command)  # ty: ignore[invalid-argument-type]

    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    @abstractmethod
    def OnMessage(
        self,
        command: Message,
    ) -> str | None:
        """Process a Message command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnCall(
        self,
        command: Call,
    ) -> str | None:
        """Process a Call command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnExecute(
        self,
        command: Execute,
    ) -> str | None:
        """Process an Execute command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnSet(
        self,
        command: Set,
    ) -> str | None:
        """Process a Set command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnAugment(
        self,
        command: Augment,
    ) -> str | None:
        """Process an Augment command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnExit(
        self,
        command: Exit,
    ) -> str | None:
        """Process an Exit command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnExitOnError(
        self,
        command: ExitOnError,
    ) -> str | None:
        """Process an ExitOnError command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnEchoOff(
        self,
        command: EchoOff,
    ) -> str | None:
        """Process an EchoOff command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnPersistError(
        self,
        command: PersistError,
    ) -> str | None:
        """Process a PersistError command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnPushDirectory(
        self,
        command: PushDirectory,
    ) -> str | None:
        """Process a PushDirectory command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnPopDirectory(
        self,
        command: PopDirectory,
    ) -> str | None:
        """Process a PopDirectory command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnRaw(
        self,
        command: Raw,
    ) -> str | None:
        """Process a Raw command."""
        raise Exception("Abstract method")  # noqa: EM101, TRY003  # pragma: no cover



# noqa: D100
import textwrap

from dbrownell_Common.Types import override

from dbrownell_ToolsDirectory.Shell.Commands import (
    Message,
    Call,
    Execute,
    Set,
    Augment,
    Exit,
    ExitOnError,
    EchoOff,
    PersistError,
    PushDirectory,
    PopDirectory,
    Raw,
)
from dbrownell_ToolsDirectory.Shell.CommandVisitor import CommandVisitor


# ----------------------------------------------------------------------
class PowerShellCommandVisitor(CommandVisitor):
    """Command visitor for PowerShell scripts."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__()

        self._message_substitution_lookup: dict[str, str] = {
            "`": "``",
            "$": "`$",
            '"': '`"',
        }

    # ----------------------------------------------------------------------
    @override
    def OnMessage(  # noqa: D102
        self,
        command: Message,
    ) -> str | None:
        output: list[str] = []

        for command_line in command.value.split("\n"):
            if not command_line.strip():
                output.append('Write-Host ""')
                continue

            line = command_line

            for source, dest in self._message_substitution_lookup.items():
                line = line.replace(source, dest)

            output.append(f'Write-Host "{line}"')

        return "; ".join(output) + "\n"

    # ----------------------------------------------------------------------
    @override
    def OnCall(  # noqa: D102
        self,
        command: Call,
    ) -> str | None:
        result = f". {command.command_line}\n"
        if command.exit_on_error:
            exit_on_error_result = self.Accept(
                ExitOnError(use_return_statement=command.exit_via_return_statement)
            )

            if exit_on_error_result:
                result += exit_on_error_result

        return result

    # ----------------------------------------------------------------------
    @override
    def OnExecute(  # noqa: D102
        self,
        command: Execute,
    ) -> str | None:
        result = f"& {command.command_line}\n"

        if command.exit_on_error:
            exit_on_error_result = self.Accept(
                ExitOnError(use_return_statement=command.exit_via_return_statement)
            )

            if exit_on_error_result:
                result += exit_on_error_result

        return result

    # ----------------------------------------------------------------------
    @override
    def OnSet(  # noqa: D102
        self,
        command: Set,
    ) -> str | None:
        if command.value_or_values is None:
            return f"Remove-Item Env:{command.name} -ErrorAction SilentlyContinue\n"

        values = ";".join(command.EnumValues())

        values = values.removeprefix('"')
        values = values.removesuffix('"')

        return f'$env:{command.name} = "{values}"\n'

    # ----------------------------------------------------------------------
    @override
    def OnAugment(  # noqa: D102
        self,
        command: Augment,
    ) -> str | None:
        statements: list[str] = []

        for value in command.EnumValues():
            if command.append_values:
                add_statement = f'$env:{command.name} = "$env:{command.name};{value}"'
            else:
                add_statement = f'$env:{command.name} = "{value};$env:{command.name}"'

            statement = textwrap.dedent(
                f"""\
                if (-not $env:{command.name} -or ";$env:{command.name};" -notlike "*;{value};*") {{
                    {add_statement}
                }}
                """,
            )

            statements.append(statement)

        return "".join(statements)

    # ----------------------------------------------------------------------
    @override
    def OnExit(  # noqa: D102
        self,
        command: Exit,
    ) -> str | None:
        return textwrap.dedent(
            """\
            {success}
            {error}
            exit {return_code}
            """,
        ).format(
            success=textwrap.dedent(
                """\
                if ($LASTEXITCODE -eq 0) {
                    Read-Host "Press [Enter] to continue"
                }
                """,
            ).rstrip()
            if command.pause_on_success
            else "",
            error=textwrap.dedent(
                """\
                if ($LASTEXITCODE -ne 0) {
                    Read-Host "Press [Enter] to continue"
                }
                """,
            ).rstrip()
            if command.pause_on_error
            else "",
            return_code=command.return_code or 0,
        )

    # ----------------------------------------------------------------------
    @override
    def OnExitOnError(  # noqa: D102
        self,
        command: ExitOnError,
    ) -> str | None:
        variable_name = f"${command.variable_name}" if command.variable_name else "$LASTEXITCODE"
        exit_keyword = "return" if command.use_return_statement else "exit"
        return_code = command.return_code if command.return_code is not None else "$error_code"

        return textwrap.dedent(
            f"""\
            $error_code = {variable_name}
            if ($error_code -ne 0) {{
                {exit_keyword} {return_code}
            }}
            """,
        )

    # ----------------------------------------------------------------------
    @override
    def OnEchoOff(  # noqa: D102
        self,
        command: EchoOff,  # noqa: ARG002
    ) -> str | None:
        # PowerShell doesn't echo commands by default
        return None

    # ----------------------------------------------------------------------
    @override
    def OnPersistError(  # noqa: D102
        self,
        command: PersistError,
    ) -> str | None:
        return f"${command.variable_name} = $LASTEXITCODE\n"

    # ----------------------------------------------------------------------
    @override
    def OnPushDirectory(  # noqa: D102
        self,
        command: PushDirectory,
    ) -> str | None:
        if command.value is None:
            return "Push-Location $PSScriptRoot\n"

        return f'Push-Location "{command.value}"\n'

    # ----------------------------------------------------------------------
    @override
    def OnPopDirectory(  # noqa: D102
        self,
        command: PopDirectory,  # noqa: ARG002
    ) -> str | None:
        return "Pop-Location\n"

    # ----------------------------------------------------------------------
    @override
    def OnRaw(  # noqa: D102
        self,
        command: Raw,
    ) -> str | None:
        return command.value.removesuffix("\n") + "\n"
